/*
								+----------------------------------+
								|                                  |
								|    ***  Bit array class   ***    |
								|                                  |
								|   Copyright  -tHE SWINe- 2005   |
								|                                  |
								|           BitArray.cpp           |
								|                                  |
								+----------------------------------+
*/

#include <crtdbg.h>
#include <string.h>
#include <stdio.h>
#include "BitArray.h"

/*
 *								=== CBit ===
 */

/*bool CBit::operator =(bool b_value)
{
	m_p_array->Set_Bit(m_n_my_index, b_value);
	return m_p_array->b_Get_Bit(m_n_my_index);
}

CBit::operator bool() const
{
	return m_p_array->b_Get_Bit(m_n_my_index);
}*/ // in header, inline

/*
 *								=== ~CBit ===
 */

/*
 *								=== CBitArray ===
 */

/*
 *	CBitArray::CBitArray()
 *		- default constructor
 */
CBitArray::CBitArray()
	:m_p_array(0),
	m_n_array_size(0),
	m_n_array_used(0)
{
	// set array empty
}

/*
 *	CBitArray::CBitArray(int n_length)
 *		- constructor with array length specifier
 *		- query CBitArray::n_Length() before use to see if there was enough memory
 */
CBitArray::CBitArray(int n_length)
	:m_p_array(0),
	m_n_array_size(0),
	m_n_array_used(0)
{
	Alloc(n_length);
	// set array empty and alloc
}

/*
 *	CBitArray::~CBitArray()
 *		- default destructor
 */
CBitArray::~CBitArray()
{
	Free();
}

/*
 *	bool CBitArray::Alloc(int n_length)
 *		- alloc array to lenght n_length bits
 */
bool CBitArray::Alloc(int n_length)
{
	Free();

	m_n_array_size = n_length / 32 + ((n_length & 31)? 1 : 0);
	// calc necessary buffer length

	if(!(m_p_array = new unsigned int[m_n_array_size])) {
		m_n_array_size = 0;

		return false;
	}
	// alloc

	m_n_array_used = n_length;
	// set length in bits

	return true;
}

/*
 *	bool CBitArray::Extend(int n_length)
 *		- extend array to lenght n_length bits
 */
bool CBitArray::Extend(int n_length)
{
	if(!m_n_array_size)
		return Alloc(n_length);
	// use Alloc() for empty array

	int n_new_size = n_length / 32 + ((n_length & 31)? 1 : 0);
	// calc new length

	if(n_new_size <= m_n_array_size) {
		m_n_array_used = n_length;

		return true;
	}
	// if it's equally long (or even shorter), we don't have to do anything

	unsigned int *p_array;
	if(!(p_array = new unsigned int[n_new_size]))
		return false;
	// alloc new buffer

	memcpy(p_array, m_p_array, m_n_array_size * sizeof(unsigned int));
	// copy contents

	delete m_p_array;
	m_p_array = p_array;
	m_n_array_size = n_new_size;
	m_n_array_used = n_length;
	// delete old buffer

	return true;
}

/*
 *	bool CBitArray::Free()
 *		- empty array, free it's memory
 */
bool CBitArray::Free()
{
	m_n_array_size = 0;
	m_n_array_used = 0;
	if(!m_p_array)
		return false;
	delete[] m_p_array;
	m_p_array = 0;
	// delete array, set to zero

	return true;
}

/*
 *	int CBitArray::n_Length() const
 *		- return
 */
int CBitArray::n_Length() const
{
	return m_n_array_used;
}

/*
 *	const CBit CBitArray::operator [](int n_index)
 *		- user friendly [] bit access
 */
CBit CBitArray::operator [](int n_index)
{
	return CBit(this, n_index);
}

/*
 *	const CBit CBitArray::operator [](int n_index) const
 *		- user friendly [] bit access
 *		- (version for const CBitArray)
 */
const CBit CBitArray::operator [](int n_index) const
{
	return CBit((CBitArray*)this, n_index);
	// hack! does any other c++ compiller allow type cast
	// from 'const CBitArray*' to 'CBitArray*' ?
}

/*
 *	bool CBitArray::operator =(bool b_value)
 *		- set all bits in array to b_value
 *		- return b_value (so you can write
 *		  arr_a = arr_b = 1 and no array copying will occur)
 */
bool CBitArray::operator =(bool b_value)
{
	Set_Array(b_value);
	return b_value;
}

/*
 *	bool CBitArray::operator =(const CBitArray &r_array)
 *		- copy bit array r_array to this
 *		- return true if there was enough memory to hold a new array
 */
bool CBitArray::operator =(const CBitArray &r_array)
{
	if(!Extend(r_array.n_Length()))
		return false;
	memcpy(m_p_array, r_array.m_p_array, m_n_array_size);
	// alloc and copy

	return true;
}

/*
 *	unsigned int CBitArray::pow2(int x)
 *		- power of two function
 *		- positive arguments only!
 */
unsigned int CBitArray::pow2(int x)
{
	_ASSERTE(x >= 0);

	unsigned int y = 1;
	for(;x > 0; x --)
		y *= 2;
	return y;
}

/*
 *	bool CBitArray::operator +=(const CBitArray &r_array)
 *		- array concatenation
 *		- return true if there was enough memory
 *		- possibly not portable because of it's endianness
 */
bool CBitArray::operator +=(const CBitArray &r_array)
{
	int n_last_bit;
	if(!Extend((n_last_bit = n_Length()) + r_array.n_Length()))
		return false;
	// alloc

	/*
	xxxxyyyyzz  |   |   |
              aaaabbbbcccc


	array:
		alligned
		unaligned
	*/

	for(int i = 0; i < r_array.m_n_array_used;) {
		if(!(n_last_bit % 32)) { // dest array aligned
			if(!(i % 32)) { // source array aligned
				// aaaa
				//     xxxx

				m_p_array[n_last_bit >> 5] = r_array.m_p_array[i >> 5];
				i += 32;
				n_last_bit += 32;
			} else { // source array un-aligned
				// aaaa
				//   xxxx

				int n_bits = 32 - i % 32;
				// no. of bits to copy

				m_p_array[n_last_bit >> 5] = r_array.m_p_array[i >> 5] >> (i % 32);
				i += n_bits;
				n_last_bit += n_bits;
			}
		} else { // dest array un-aligned
			if(!(i % 32)) { // source array aligned
				// aa
				//   xxxx

				int n_bits = 32 - n_last_bit % 32;
				// no. of bits to copy

				m_p_array[n_last_bit >> 5] &= pow2(32 - n_bits) - 1;
				// mask out dest array

				m_p_array[n_last_bit >> 5] |= r_array.m_p_array[i >> 5] << (n_last_bit % 32);

				i += n_bits;
				n_last_bit += n_bits;
			} else { // source array un-aligned
				// aa
				// xxxx

				int n_bits = 32 - (((i % 32) < (n_last_bit % 32))? (n_last_bit % 32) : (i % 32));
				// no. of bits to copy

				m_p_array[n_last_bit >> 5] &= pow2(n_last_bit % 32) - 1;
				// mask out dest array

				m_p_array[n_last_bit >> 5] |= (r_array.m_p_array[i >> 5] >> (i % 32)) << (n_last_bit % 32);

				i += n_bits;
				n_last_bit += n_bits;
			}
		}
	}

	return true;
}

/*
 *	bool CBitArray::operator ==(bool b_value)
 *		- array / number comparison
 *		- return true (all bits in array are equal to b_value) or false
 *		- !!always!! return true for an empty array
 */
bool CBitArray::operator ==(bool b_value)
{
	for(int i = 0; i < m_n_array_used >> 5; i ++) {
		if(m_p_array[i] != (unsigned)((b_value)? -1 : 0)) // -1 = all ones in binary
			return false;
	}
	if(m_n_array_used % 32) {
		int m = pow2(m_n_array_used % 32) - 1; // mask for remaining bits
		if((m_p_array[m_n_array_used >> 5] & m) != (unsigned)((-1 * b_value) & m))
			return false;
	}

	return true;
}

/*
 *	int CBitArray::operator ==(const CBitArray &r_array)
 *		- array comparison
 *		- compare bits in order they go in the arrays
 *		- if all bits are equal, the longer array is greater
 *		- return 1 if this > r_array, 0 if this == r_array or -1 if this < r_array
 *		- possibly not portable because of it's endianness
 */
int CBitArray::operator ==(const CBitArray &r_array)
{
	int n = (m_n_array_used > r_array.m_n_array_used)? r_array.m_n_array_used : m_n_array_used;
	for(int i = 0; i < n >> 5; i ++) {
		if(m_p_array[i] != r_array.m_p_array[i])
			return (r_array.m_p_array[i] > m_p_array[i])? -1 : 1;
	}
	if(n % 32) {
		int m = pow2(n % 32) - 1; // mask for remaining bits
		if((m_p_array[n >> 5] & m) != (r_array.m_p_array[n >> 5] & m))
			return ((r_array.m_p_array[n >> 5] & m) > (m_p_array[n >> 5] & m))? -1 : 1;
	}

	return (r_array.m_n_array_used > m_n_array_used)? -1 :
		((m_n_array_used == r_array.m_n_array_used)? 0 : 1);
}

/*
 *	void CBitArray::Set_Array(bool b_value)
 *		- set all bits in array to b_value
 */
void CBitArray::Set_Array(bool b_value)
{
	if(!m_n_array_used)
		return;
	// only for non-empty arrays

	if(b_value)
		memset(m_p_array, 0xff, m_n_array_size * sizeof(int));
	else
		memset(m_p_array, 0x00, m_n_array_size * sizeof(int));
	// set whole array to 0-s or 1-s, depending on b_value
}

/*
 *	void CBitArray::Set_Bit(int n_index, bool b_value)
 *		- set bit with index n_index to value b_value
 */
void CBitArray::Set_Bit(int n_index, bool b_value)
{
	if(b_value)
		Set_Bit_High(n_index);
	else
		Set_Bit_Low(n_index);
	// set bit with index n_index to b_value
}

/*
 *	bool CBitArray::b_Get_Bit(int n_index) const
 *		- return value of bit with index n_index
 *		- no bad index checking
 */
bool CBitArray::b_Get_Bit(int n_index) const
{
	/*if(n_index > m_n_array_used || n_index < 0)
		return -1;*/

	return (m_p_array[n_index >> 5] >> (n_index & 31)) & 1;
	// get bit
}

/*
 *	void CBitArray::Set_Bit_High(int n_index)
 *		- set bit with index n_index to 1
 */
void CBitArray::Set_Bit_High(int n_index)
{
	/*if(n_index > m_n_array_used || n_index < 0)
		return;*/

	m_p_array[n_index >> 5] |= 1 << (n_index & 31);
	// set bit to 1
}

/*
 *	void CBitArray::Set_Bit_Low(int n_index)
 *		- set bit with index n_index to 0
 */
void CBitArray::Set_Bit_Low(int n_index)
{
	/*if(n_index > m_n_array_used || n_index < 0)
		return;*/

	m_p_array[n_index >> 5] &= ~(1 << (n_index & 31));
	// set bit to 0
}

/*
 *	bool CBitArray::Store(FILE *p_fw) const
 *		- write an array to file p_fw (binary)
 *		- return true if everything was ok (enough space for writing, etc ...)
 */
bool CBitArray::Store(FILE *p_fw) const
{
	if(fwrite(&m_n_array_used, sizeof(int), 1, p_fw) != 1 ||
	   fwrite(m_p_array, sizeof(int), m_n_array_size, p_fw) != (unsigned)m_n_array_size)
		return false;

	return true;
}

/*
 *	bool CBitArray::Load(FILE *p_fr)
 *		- load an array from file p_fr (binary)
 *		- return true when reading went ok and there was enough memory to hold a new array
 */
bool CBitArray::Load(FILE *p_fr)
{
	Free();

	if(fread(&m_n_array_used, sizeof(int), 1, p_fr) != 1)
		return false;
	m_n_array_size = m_n_array_used / 32 + ((m_n_array_used & 31)? 1 : 0);
	if(!(m_p_array = new unsigned int[m_n_array_size])) {
		m_n_array_size = 0;

		return false;
	}
	if(fread(m_p_array, sizeof(int), m_n_array_size, p_fr) != (unsigned)m_n_array_size)
		return false;

	return true;
}

/*
 *								=== ~CBitArray ===
 */
